from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from influxdb_client import Point
from pydantic import Field, computed_field
from pydantic.json_schema import SkipJsonSchema

from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.models import Solaredge2MQTTBaseModel
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantSensorType as HASensor,
)
from solaredge2mqtt.services.modbus.models.base import ModbusUnitInfo, ModbusUnitRole
from solaredge2mqtt.services.modbus.models.battery import ModbusBattery
from solaredge2mqtt.services.modbus.models.inverter import ModbusInverter
from solaredge2mqtt.services.modbus.models.meter import ModbusMeter
from solaredge2mqtt.services.modbus.models.unit import ModbusUnit
from solaredge2mqtt.services.models import Component

if TYPE_CHECKING:
    from solaredge2mqtt.services.energy.settings import PriceSettings


class Powerflow(Component):
    COMPONENT = "powerflow"
    SOURCE = None

    unit: ModbusUnitInfo | None = Field(None, title="Unit Info")

    pv_production: int = Field(0, **HASensor.POWER_W.field("PV production"))
    inverter: InverterPowerflow = Field(title="Inverter")
    grid: GridPowerflow = Field(title="Grid")
    battery: BatteryPowerflow = Field(title="Battery")
    consumer: ConsumerPowerflow = Field(title="Consumer")

    last_powerflow: ClassVar[Powerflow] = None

    @staticmethod
    def from_modbus(
        unit: ModbusUnit,
        evcharger: int = 0,
    ) -> Powerflow:
        inverter_data = unit.inverter

        grid = GridPowerflow.from_modbus(unit.meters)
        battery = BatteryPowerflow.from_modbus(unit.batteries)

        if inverter_data.ac.power.actual > 0:
            pv_production = int(inverter_data.dc.power + battery.power)
            if pv_production < 0:
                pv_production = 0
        else:
            pv_production = 0

        inverter = InverterPowerflow.from_modbus(inverter_data, battery)

        consumer = ConsumerPowerflow(inverter, grid, evcharger)

        return Powerflow(
            unit=unit.info,
            pv_production=pv_production,
            inverter=inverter,
            grid=grid,
            battery=battery,
            consumer=consumer,
        )

    @staticmethod
    def cumulated_powerflow(powerflows: dict[str, Powerflow]) -> Powerflow:
        unit = ModbusUnitInfo(
            key="cumulated",
            unit=0,
            role=ModbusUnitRole.CUMULATED,
        )

        inverter = InverterPowerflow(
            power=sum(p.inverter.power for p in powerflows.values()),
            dc_power=sum(p.inverter.dc_power for p in powerflows.values()),
            battery_discharge=sum(
                p.inverter.battery_discharge for p in powerflows.values()),
        )

        grid = GridPowerflow(
            power=sum(p.grid.power for p in powerflows.values()))
        battery = BatteryPowerflow(
            power=sum(p.battery.power for p in powerflows.values()))

        consumer = ConsumerPowerflow(
            inverter=inverter,
            grid=grid,
            evcharger=sum(p.consumer.evcharger for p in powerflows.values()),
        )

        pv_production = sum(p.pv_production for p in powerflows.values())

        return Powerflow(
            unit=unit,
            pv_production=pv_production,
            inverter=inverter,
            grid=grid,
            battery=battery,
            consumer=consumer,
        )

    def is_valid(self, external_production: bool) -> bool:
        valid = False

        if self.pv_production < 0:
            logger.warning("PV production is negative")
        elif (
            not external_production and
            self.consumer.used_production + self.grid.delivery
            != self.inverter.production
        ):
            logger.warning(
                (
                    "Consumer used production + grid delivery "
                    "is not equal to inverter production"
                )
            )
            valid = True
        else:
            valid = all(
                [
                    self.inverter.is_valid,
                    self.grid.is_valid,
                    self.battery.is_valid,
                    self.consumer.is_valid,
                ]
            )

        return valid

    @classmethod
    def is_not_valid_with_last(cls, powerflow: Powerflow) -> bool:
        check = False

        if cls.last_powerflow is not None:
            check = all(
                [
                    cls.last_powerflow.pv_production == 0
                    and powerflow.pv_production > 100,
                ]
            )

        cls.last_powerflow = powerflow

        return check

    def prepare_point(self, measurement: str = "powerflow_raw") -> Point:
        point = Point(measurement)
        for key, value in self.model_dump_influxdb(exclude=["unit"]).items():
            point.field(key, value)

        if self.has_unit:
            point.tag("unit", self.unit.key)

        return point

    def prepare_point_energy(
        self, measurement: str = "energy", prices: PriceSettings = None
    ) -> Point:
        point = Point(measurement)
        for key, value in self.model_dump_influxdb().items():
            energy = value / 1000
            point.field(key, energy)
            if prices is not None:
                if key == "consumer_used_production":
                    point.field("money_saved", energy * prices.price_in)
                    point.field("money_price_in", prices.price_in)
                elif key == "grid_delivery":
                    point.field("money_delivered", energy * prices.price_out)
                    point.field("money_price_out", prices.price_out)
                elif key == "grid_consumption":
                    point.field("money_consumed", energy * prices.price_in)

        return point

    @property
    def has_unit(self) -> bool:
        return self.unit is not None

    def mqtt_topic(self) -> str:
        topic = super().mqtt_topic()

        if self.has_unit:
            topic += f"/{self.unit.key.lower()}"

        return topic

    def homeassistant_device_info(self) -> dict[str, any]:
        name = "Powerflow"

        if self.has_unit:
            name += f" {self.unit.key}"

        return self._default_homeassistant_device_info(name)


class InverterPowerflow(Solaredge2MQTTBaseModel):
    power: SkipJsonSchema[int]
    dc_power: int = Field(
        **HASensor.POWER_W.field("Power DC", "solar-power"))
    battery_discharge: SkipJsonSchema[int] = Field(exclude=True)

    @staticmethod
    def from_modbus(
        inverter_data: ModbusInverter,
        battery: BatteryPowerflow,
    ) -> InverterPowerflow:
        power = int(inverter_data.ac.power.actual)
        dc_power = int(inverter_data.dc.power)

        return InverterPowerflow(
            power=power,
            dc_power=dc_power,
            battery_discharge=battery.discharge,
        )

    @property
    def battery_factor(self) -> float:
        factor = 0.0
        if self.power > 0 and self.battery_discharge > 0:
            factor = self.battery_discharge / self.dc_power

        return factor

    @computed_field(**HASensor.POWER_W.field("Consumption"))
    @property
    def consumption(self) -> int:
        return abs(self.power) if self.power < 0 else 0

    @computed_field(**HASensor.POWER_W.field("Production"))
    @property
    def production(self) -> int:
        return self.power if self.power > 0 else 0

    @computed_field(
        **HASensor.POWER_W.field("Battery production", "home-battery-outline")
    )
    @property
    def battery_production(self) -> int:
        battery_production = 0
        if self.production > 0 and self.battery_factor > 0:
            battery_production = int(
                round(self.production * self.battery_factor))
            battery_production = min(battery_production, self.production)
        return battery_production

    @computed_field(**HASensor.POWER_W.field("PV production", "sun-angle-outline"))
    @property
    def pv_production(self) -> int:
        return self.production - self.battery_production

    @property
    def is_valid(self) -> bool:
        valid = False
        if self.consumption < 0:
            logger.warning("Inverter consumption is negative")
        elif self.production < 0:
            logger.warning("Inverter production is negative")
        elif self.pv_production < 0:
            logger.warning("Inverter PV production is negative")
        elif self.battery_production < 0:
            logger.warning("Inverter battery production is negative")
        else:
            valid = True

        return valid


class GridPowerflow(Solaredge2MQTTBaseModel):
    power: SkipJsonSchema[int]

    @staticmethod
    def from_modbus(meters_data: dict[str, ModbusMeter]) -> GridPowerflow:
        grid = 0
        for meter in meters_data.values():
            if "Import" in meter.info.option and "Export" in meter.info.option:
                grid += meter.power.actual

        return GridPowerflow(power=round(grid))

    @computed_field(
        **HASensor.POWER_W.field("Consumption", "transmission-tower-import")
    )
    @property
    def consumption(self) -> int:
        return abs(self.power) if self.power < 0 else 0

    @computed_field(**HASensor.POWER_W.field("Delivery", "transmission-tower-export"))
    @property
    def delivery(self) -> int:
        return self.power if self.power > 0 else 0

    @property
    def is_valid(self) -> bool:
        valid = False
        if self.consumption < 0:
            logger.warning("Grid consumption is negative")
        elif self.delivery < 0:
            logger.warning("Grid delivery is negative")
        else:
            valid = True

        return valid


class BatteryPowerflow(Solaredge2MQTTBaseModel):
    power: SkipJsonSchema[int]

    @staticmethod
    def from_modbus(batteries_data: dict[str, ModbusBattery]) -> BatteryPowerflow:
        batteries_power = 0
        for battery in batteries_data.values():
            batteries_power += battery.power

        return BatteryPowerflow(power=batteries_power)

    @computed_field(**HASensor.POWER_W.field("Charge", "battery-plus-outline"))
    @property
    def charge(self) -> int:
        return self.power if self.power > 0 else 0

    @computed_field(**HASensor.POWER_W.field("Discharge", "battery-minus-outline"))
    @property
    def discharge(self) -> int:
        return abs(self.power) if self.power < 0 else 0

    @property
    def is_valid(self) -> bool:
        valid = False
        if self.charge < 0:
            logger.warning("Battery charge is negative")
        elif self.discharge < 0:
            logger.warning("Battery discharge is negative")
        else:
            valid = True

        return valid


class ConsumerPowerflow(Solaredge2MQTTBaseModel):
    house: int = Field(
        **HASensor.POWER_W.field("House", "home-lightning-bolt-outline")
    )

    evcharger: int = Field(
        0, **HASensor.POWER_W.field("EV-Charger", "ev-station"))

    inverter: int = Field(**HASensor.POWER_W.field("Inverter"))

    used_production: int = Field(
        0, **HASensor.POWER_W.field("Used production"))

    battery_factor: SkipJsonSchema[float] = Field(exclude=True)

    def __init__(
        self, inverter: InverterPowerflow, grid: GridPowerflow, evcharger: int
    ):
        house = int(abs(grid.power - inverter.power)) - evcharger

        battery_factor = inverter.battery_factor

        if inverter.production > 0 and inverter.production > grid.delivery:
            used_production = inverter.production - grid.delivery
        else:
            used_production = 0

        super().__init__(
            house=house,
            evcharger=evcharger,
            inverter=inverter.consumption,
            used_production=used_production,
            battery_factor=battery_factor,
        )

    @computed_field(**HASensor.POWER_W.field("Total"))
    @property
    def total(self) -> int:
        return self.house + self.evcharger + self.inverter

    @computed_field(
        **HASensor.POWER_W.field("Used consumption"),
    )
    @property
    def used_battery_production(self) -> int:
        battery_production = 0
        if self.used_production > 0 and self.battery_factor > 0:
            battery_production = int(
                round(self.used_production * self.battery_factor))
            battery_production = min(battery_production, self.used_production)
        return battery_production

    @computed_field(
        **HASensor.POWER_W.field("Used PV production"),
    )
    @property
    def used_pv_production(self) -> int:
        return self.used_production - self.used_battery_production

    @property
    def is_valid(self) -> bool:
        valid = False
        if self.house < 0:
            logger.warning("Consumer house is negative")
        elif self.evcharger < 0:
            logger.warning("Consumer evcharger is negative")
        elif self.inverter < 0:
            logger.warning("Consumer inverter is negative")
        elif self.used_production < 0:
            logger.warning("Consumer used production is negative")
        elif self.used_pv_production < 0:
            logger.warning("Consumer used PV production is negative")
        elif self.used_battery_production < 0:
            logger.warning("Consumer used battery production is negative")
        elif self.total < 0:
            logger.warning("Consumer total is negative")
        elif self.total < self.used_production:
            logger.warning("Consumer total is less than used production")
        else:
            valid = True

        return valid



from pydantic import BaseModel, Field, SecretStr


class MQTTSettings(BaseModel):
    client_id: str = Field("solaredge2mqtt")
    broker: str
    port: int = Field(1883)
    username: str | None = Field(None)
    password: SecretStr | None = Field(None)
    topic_prefix: str = Field("solaredge")

    @property
    def kargs(self) -> dict[str, str]:
        args = {"identifier": self.client_id}

        if self.username is not None:
            args["username"] = self.username
        if self.password is not None:
            args["password"] = self.password.get_secret_value()

        return args



from __future__ import annotations

from typing import TYPE_CHECKING

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import (
    MQTTPublishEvent,
    MQTTReceivedEvent,
    MQTTSubscribeEvent,
)
from solaredge2mqtt.services.energy.events import EnergyReadEvent
from solaredge2mqtt.services.events import ComponentEvent
from solaredge2mqtt.services.forecast.events import ForecastEvent
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantBinarySensorType,
    HomeAssistantDevice,
    HomeAssistantEntity,
    HomeAssistantNumberType,
    HomeAssistantSensorType,
    HomeAssistantStatus,
    HomeAssistantStatusInput,
    HomeAssistantType,
)
from solaredge2mqtt.services.modbus.events import ModbusUnitsReadEvent
from solaredge2mqtt.services.modbus.models.inverter import ModbusInverter
from solaredge2mqtt.services.models import Component
from solaredge2mqtt.services.powerflow.events import PowerflowGeneratedEvent
from solaredge2mqtt.services.wallbox.events import WallboxReadEvent

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings import ServiceSettings


class HomeAssistantDiscovery:
    def __init__(self, service_settings: ServiceSettings, event_bus: EventBus) -> None:
        self.settings = service_settings
        self._send_entities: dict[str, HomeAssistantEntity] = {}
        logger.info("Home Assistant discovery enabled")

        self._status_topic = f"{self.settings.homeassistant.topic_prefix}/status"

        self._seen_energy_periods: set[str] = set()

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(
            [
                ForecastEvent,
                EnergyReadEvent,
                WallboxReadEvent,
            ],
            self.component_discovery,
        )

        self.event_bus.subscribe(
            PowerflowGeneratedEvent, self.powerflow_discovery
        )

        self.event_bus.subscribe(
            ModbusUnitsReadEvent, self.units_discovery
        )

        self.event_bus.subscribe(
            MQTTReceivedEvent,
            self.homeassistant_status,
        )

    async def async_init(self) -> None:
        await self.event_bus.emit(
            MQTTSubscribeEvent(
                self._status_topic,
                HomeAssistantStatusInput,
            )
        )

    async def component_discovery(self, event: ComponentEvent) -> None:
        publish = True
        if isinstance(event, EnergyReadEvent):
            period = event.component.info.period
            publish = (
                period.auto_discovery and (
                    event.component.mqtt_topic() not in self._seen_energy_periods
                )
            )
            self._seen_energy_periods.add(event.component.mqtt_topic())
        else:
            self.event_bus.unsubscribe(event, self.component_discovery)

        if publish:
            logger.info(
                f"Home Assistant discovery component: {event.component}")
            device_info = event.component.homeassistant_device_info()
            state_topic = self.state_topic(event.component.mqtt_topic())
            await self.publish_component(event.component, device_info, state_topic)

    async def units_discovery(self, event: ModbusUnitsReadEvent) -> None:
        self.event_bus.unsubscribe(event, self.units_discovery)
        for unit_key, unit in event.units.items():
            logger.info(f"Home Assistant discovery {unit_key}:inverter")

            device_info = unit.inverter.homeassistant_device_info()
            state_topic = self.state_topic(
                unit.inverter.mqtt_topic(self.settings.modbus.has_followers))
            await self.publish_component(unit.inverter, device_info, state_topic)

            for name, component in {**unit.meters, **unit.batteries}.items():
                logger.info(f"Home Assistant discovery {unit_key}:{name}")

                device_info = component.homeassistant_device_info_with_name(
                    name)
                state_topic = self.state_topic(
                    component.mqtt_topic(self.settings.modbus.has_followers), name)
                await self.publish_component(component, device_info, state_topic)

    async def powerflow_discovery(self, event: PowerflowGeneratedEvent) -> None:
        self.event_bus.unsubscribe(event, self.powerflow_discovery)

        for key, powerflow in event.components.items():
            logger.info(f"Home Assistant discovery {key}:powerflow")

            device_info = powerflow.homeassistant_device_info()
            state_topic = self.state_topic(powerflow.mqtt_topic())
            await self.publish_component(powerflow, device_info, state_topic)

    def state_topic(self, component_topic: str, name: str = "") -> str:
        subtopic = f"/{name.lower()}" if name else ""
        return f"{self.settings.mqtt.topic_prefix}/{component_topic}{subtopic}"

    async def publish_component(
        self,
        component: Component,
        device_info: dict[str, any],
        state_topic: str
    ) -> None:
        logger.trace(device_info)
        device = HomeAssistantDevice(
            client_id=self.settings.mqtt.client_id,
            state_topic=state_topic,
            **device_info,
        )
        logger.debug(device)

        entities_info = component.parse_schema(self.property_parser)
        for entity_info in entities_info:
            if isinstance(component, ModbusInverter):
                path = entity_info["path"]
                if (
                    not self.settings.modbus.check_grid_status
                    and path[0] == "grid_status"
                ):
                    continue

                if (
                    not self.settings.modbus.advanced_power_controls_enabled
                    and path[0] == "advanced_power_controls"
                ):
                    continue

            if (
                self.settings.is_prices_configured
                and entity_info["ha_type"] == HomeAssistantSensorType.MONETARY
            ):
                entity_info["unit"] = self.settings.prices.currency

            entity = HomeAssistantEntity(
                device=device,
                **entity_info,
            )
            topic = (
                f"{entity.ha_type.typed}/"
                f"{state_topic.replace('/', '_')}/"
                f"{entity.unique_id}/config"
            )
            logger.debug(entity)

            self._send_entities[topic] = entity

            await self.event_bus.emit(
                MQTTPublishEvent(
                    topic=topic,
                    payload=entity,
                    retain=self.settings.homeassistant.retain,
                    topic_prefix=self.settings.homeassistant.topic_prefix,
                    exclude_none=True,
                )
            )

    async def homeassistant_status(self, event: MQTTReceivedEvent) -> None:
        if event.topic == self._status_topic:
            status = event.input.status
            if status == HomeAssistantStatus.ONLINE:
                logger.info(
                    "Home Assistant status changed to online resend discovery")
                for topic, entity in self._send_entities.items():
                    await self.event_bus.emit(
                        MQTTPublishEvent(
                            topic=topic,
                            payload=entity,
                            retain=self.settings.homeassistant.retain,
                            topic_prefix=self.settings.homeassistant.topic_prefix,
                            exclude_none=True,
                        )
                    )

    @staticmethod
    def property_parser(prop, name: str, path: list[str]) -> dict | None:
        entity: dict | None = None

        if "ha_type" in prop:
            entity = {"name": name,
                      "path": path}

            ha_type = prop["ha_type"]
            entity["icon"] = prop["icon"]

            typed = HomeAssistantType.from_string(
                prop["ha_typed"]
            )

            if typed == HomeAssistantType.BINARY_SENSOR:
                entity["ha_type"] = HomeAssistantBinarySensorType.from_string(
                    ha_type)
            elif typed == HomeAssistantType.NUMBER:
                entity["ha_type"] = HomeAssistantNumberType.from_string(
                    ha_type)
            elif typed == HomeAssistantType.SENSOR:
                entity["ha_type"] = HomeAssistantSensorType.from_string(
                    ha_type)

            for field in typed.additional_fields:
                entity[field] = prop.get(field, None)

        return entity



from solaredge2mqtt.core.events.events import BaseEvent
from solaredge2mqtt.services.modbus.models.unit import ModbusUnit
from solaredge2mqtt.services.modbus.sunspec.base import SunSpecRegister
from solaredge2mqtt.services.modbus.sunspec.values import SunSpecInputData


class ModbusUnitsReadEvent(BaseEvent):
    def __init__(self, units: dict[str, ModbusUnit]):
        self._units = units

    @property
    def units(self) -> dict[str, ModbusUnit]:
        return self._units


class ModbusWriteEvent(BaseEvent):
    AWAIT = True

    def __init__(self,
                 register: SunSpecRegister,
                 payload: SunSpecInputData):
        self._register = register
        self._payload = payload

    @property
    def register(self) -> SunSpecRegister:
        return self._register

    @property
    def payload(self) -> SunSpecInputData:
        return self._payload



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from aiohttp import ClientResponseError

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.core.timer.events import Interval10MinTriggerEvent
from solaredge2mqtt.services.http_async import HTTPClientAsync
from solaredge2mqtt.services.weather.events import WeatherUpdateEvent
from solaredge2mqtt.services.weather.models import OpenWeatherMapOneCall

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings import ServiceSettings

ONECALL_URL = "https://api.openweathermap.org/data/3.0/onecall"
TIMEMACHINE_URL = "https://api.openweathermap.org/data/3.0/onecall/timemachine"


class WeatherClient(HTTPClientAsync):
    def __init__(self, settings: ServiceSettings, event_bus: EventBus) -> None:
        super().__init__("Weather API")

        self.location = settings.location
        self.settings = settings.weather

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self):
        self.event_bus.subscribe(Interval10MinTriggerEvent, self.loop)

    async def loop(self, _):
        weather = await self.get_weather()
        await self.event_bus.emit(WeatherUpdateEvent(weather))
        await self.event_bus.emit(
            MQTTPublishEvent(
                "weather/current",
                weather.current,
                self.settings.retain,
            )
        )

    async def get_weather(self) -> OpenWeatherMapOneCall:
        try:
            logger.info("Reading weather data from OpenWeatherMap")
            async with asyncio.timeout(7):
                result = await self._get(
                    ONECALL_URL,
                    params={
                        "lat": self.location.latitude,
                        "lon": self.location.longitude,
                        "exclude": "minutely,daily,alerts",
                        "units": "metric",
                        "lang": self.settings.language,
                        "appid": self.settings.api_key.get_secret_value(),
                    },
                )
            logger.trace(result)

            if result is None:
                raise InvalidDataException(
                    "Unable to read weather data from OpenWeatherMap"
                )

            weather = OpenWeatherMapOneCall(**result)

            return weather
        except ClientResponseError as error:
            status_code = error.status
            if status_code == 401:
                error_msg = (
                    "Invalid OpenWeatherMap API key or no subscription to OneCall-API"
                )
            else:
                error_msg = (
                    "Unable to read weather data from OpenWeatherMap "
                    "(HTTP {status_code})"
                )

            raise InvalidDataException(error_msg) from error
        except asyncio.TimeoutError as error:
            raise InvalidDataException(
                "Unable to read weather data from OpenWeatherMap (timeout)"
            ) from error



from solaredge2mqtt.services.modbus.sunspec.base import SunSpecRegister
from solaredge2mqtt.services.modbus.sunspec.battery import (
    SunSpecBatteryInfoRegister,
    SunSpecBatteryOffset,
)
from solaredge2mqtt.services.modbus.sunspec.meter import (
    SunSpecMeterInfoRegister,
    SunSpecMeterOffset,
)
from solaredge2mqtt.services.modbus.sunspec.values import (
    SunSpecPayload,
    SunSpecValueType,
)


class SunSpecInverterInfoRegister(SunSpecRegister):
    C_ID = "c_id", 40000, SunSpecValueType.STRING, False, 2
    C_DID = "c_did", 40002, SunSpecValueType.UINT16
    C_LENGTH = "c_length", 40003, SunSpecValueType.UINT16
    C_MANUFACTURER = ("c_manufacturer", 40004,
                      SunSpecValueType.STRING, True, 16)
    C_MODEL = "c_model", 40020, SunSpecValueType.STRING, True, 16
    C_VERSION = "c_version", 40044, SunSpecValueType.STRING, True, 8
    C_SERIALNUMBER = ("c_serialnumber", 40052,
                      SunSpecValueType.STRING, True, 16)
    C_DEVICEADDRESS = "c_deviceaddress", 40068, SunSpecValueType.UINT16, True
    C_SUNSPEC_DID = "c_sunspec_did", 40069, SunSpecValueType.UINT16, True
    C_SUNSPEC_LENGTH = "c_sunspec_length", 40070, SunSpecValueType.UINT16

    METER0 = (
        "meter0",
        SunSpecMeterInfoRegister.C_SUNSPEC_DID.address
        + SunSpecMeterOffset.METER0.offset,
        SunSpecMeterInfoRegister.C_SUNSPEC_DID.value_type,
        True,
    )
    METER1 = (
        "meter1",
        SunSpecMeterInfoRegister.C_SUNSPEC_DID.address
        + SunSpecMeterOffset.METER1.offset,
        SunSpecMeterInfoRegister.C_SUNSPEC_DID.value_type,
        True,
    )
    METER2 = (
        "meter2",
        SunSpecMeterInfoRegister.C_SUNSPEC_DID.address
        + SunSpecMeterOffset.METER2.offset,
        SunSpecMeterInfoRegister.C_SUNSPEC_DID.value_type,
        True,
    )

    BATTERY0 = (
        "battery0",
        SunSpecBatteryInfoRegister.C_DEVICEADDRESS.address
        + SunSpecBatteryOffset.BATTERY0.offset,
        SunSpecBatteryInfoRegister.C_DEVICEADDRESS.value_type,
        True,
    )

    BATTERY1 = (
        "battery1",
        SunSpecBatteryInfoRegister.C_DEVICEADDRESS.address
        + SunSpecBatteryOffset.BATTERY1.offset,
        SunSpecBatteryInfoRegister.C_DEVICEADDRESS.value_type,
        True,
    )


class SunSpecInverterRegister(SunSpecRegister):
    CURRENT = "current", 40071, SunSpecValueType.UINT16, True
    L1_CURRENT = "l1_current", 40072, SunSpecValueType.UINT16
    L2_CURRENT = "l2_current", 40073, SunSpecValueType.UINT16
    L3_CURRENT = "l3_current", 40074, SunSpecValueType.UINT16
    CURRENT_SCALE = "current_scale", 40075, SunSpecValueType.INT16, True

    L1_VOLTAGE = "l1_voltage", 40076, SunSpecValueType.UINT16, True
    L2_VOLTAGE = "l2_voltage", 40077, SunSpecValueType.UINT16, True
    L3_VOLTAGE = "l3_voltage", 40078, SunSpecValueType.UINT16, True
    L1N_VOLTAGE = "l1n_voltage", 40079, SunSpecValueType.UINT16, True
    L2N_VOLTAGE = "l2n_voltage", 40080, SunSpecValueType.UINT16, True
    L3N_VOLTAGE = "l3n_voltage", 40081, SunSpecValueType.UINT16, True
    VOLTAGE_SCALE = "voltage_scale", 40082, SunSpecValueType.INT16, True

    POWER_AC = "power_ac", 40083, SunSpecValueType.INT16, True
    POWER_AC_SCALE = "power_ac_scale", 40084, SunSpecValueType.INT16, True

    FREQUENCY = "frequency", 40085, SunSpecValueType.UINT16, True
    FREQUENCY_SCALE = "frequency_scale", 40086, SunSpecValueType.INT16, True

    POWER_APPARENT = "power_apparent", 40087, SunSpecValueType.INT16, True
    POWER_APPARENT_SCALE = "power_apparent_scale", 40088, SunSpecValueType.INT16, True

    POWER_REACTIVE = "power_reactive", 40089, SunSpecValueType.INT16, True
    POWER_REACTIVE_SCALE = (
        "power_reactive_scale",
        40090,
        SunSpecValueType.INT16,
        True,
    )

    POWER_FACTOR = "power_factor", 40091, SunSpecValueType.INT16, True
    POWER_FACTOR_SCALE = "power_factor_scale", 40092, SunSpecValueType.INT16, True

    ENERGY_TOTAL = "energy_total", 40093, SunSpecValueType.UINT32, True
    ENERGY_TOTAL_SCALE = "energy_total_scale", 40095, SunSpecValueType.INT16, True

    CURRENT_DC = "current_dc", 40096, SunSpecValueType.UINT16, True
    CURRENT_DC_SCALE = "current_dc_scale", 40097, SunSpecValueType.INT16, True

    VOLTAGE_DC = "voltage_dc", 40098, SunSpecValueType.UINT16, True
    VOLTAGE_DC_SCALE = "voltage_dc_scale", 40099, SunSpecValueType.INT16, True

    POWER_DC = "power_dc", 40100, SunSpecValueType.INT16, True
    POWER_DC_SCALE = "power_dc_scale", 40101, SunSpecValueType.INT16, True

    TEMPERATURE = "temperature", 40103, SunSpecValueType.INT16, True
    TEMPERATURE_SCALE = "temperature_scale", 40106, SunSpecValueType.INT16, True

    STATUS = "status", 40107, SunSpecValueType.UINT16, True
    VENDOR_STATUS = "vendor_status", 40108, SunSpecValueType.UINT16


class SunSpecGridStatusRegister(SunSpecRegister):
    GRID_STATUS = "grid_status", 40113, SunSpecValueType.UINT32, True


class SunSpecPowerControlRegister(SunSpecRegister):
    RRCR_STATE = "rrcr_state", 61440, SunSpecValueType.UINT16, True
    ACTIVE_POWER_LIMIT = "active_power_limit", 61441, SunSpecValueType.UINT16, True
    COSPHI = "cosphi", 61442, SunSpecValueType.FLOAT32, False

    COMMIT_POWER_CONTROL_SETTINGS = (
        "commit_power_control_settings",
        61696,
        SunSpecValueType.INT16,
        True
    )
    RESTORE_POWER_CONTROL_SETTINGS = (
        "restore_power_control_settings",
        61697,
        SunSpecValueType.INT16,
        True
    )

    ADVANCED_POWER_CONTROL_ENABLE = (
        "advanced_power_control_enable",
        61762,
        SunSpecValueType.INT32,
        True
    )

    REACTIVE_POWER_CONFIG = "reactive_power_config", 61700, SunSpecValueType.INT32, True
    REACTIVE_POWER_RESPONSE_TIME = (
        "reactive_power_response_time",
        61702,
        SunSpecValueType.UINT32, True
    )

    def decode_response(
        self, registers: list[int], data: dict[str, SunSpecPayload]
    ) -> dict[str, SunSpecPayload]:
        data = super().decode_response(registers, data)

        if self == SunSpecPowerControlRegister.ADVANCED_POWER_CONTROL_ENABLE:
            data[self.identifier] = data[self.identifier] == 1

        return data

    @staticmethod
    def wordorder() -> str:
        return "little"


class SunSpecSiteLimitRegister(SunSpecRegister):
    EXPORT_CONTROL_MODE = "export_control_mode", 57344, SunSpecValueType.UINT16, True
    EXPORT_CONTROL_LIMIT_MODE = (
        "export_control_limit_mode",
        57345,
        SunSpecValueType.UINT16,
        True
    )
    EXPORT_CONTROL_SITE_LIMIT = (
        "export_control_site_limit",
        57346,
        SunSpecValueType.FLOAT32,
        True
    )

    def decode_response(
        self, registers: list[int], data: dict[str, SunSpecPayload]
    ) -> dict[str, SunSpecPayload]:
        data = super().decode_response(registers, data)

        if self == SunSpecPowerControlRegister.EXPORT_CONTROL_SITE_LIMIT:
            data[self.identifier] = 0 if data[self.identifier] < 0 else int(
                data[self.identifier])
        elif self == SunSpecPowerControlRegister.EXPORT_CONTROL_MODE:
            bitmask = data[self.identifier]
            data[f"{self.identifier}_raw"] = bitmask

            data[self.identifier] = ((bitmask & 0x0001) +
                                     ((bitmask & 0x0002) >> 1) +
                                     ((bitmask & 0x0004) >> 2))

            data["export_control_external_production"] = bool(bitmask & 0x0400)
            data["export_control_negative_site_limit"] = bool(bitmask & 0x0800)

        return data

    @staticmethod
    def wordorder() -> str:
        return "little"



from pydantic import BaseModel, Field, SecretStr


class WallboxSettings(BaseModel):
    host: str = Field(None)
    password: SecretStr = Field(None)
    serial: str = Field(None)
    retain: bool = Field(False)

    @property
    def is_configured(self) -> bool:
        return all(
            [self.host is not None, self.password is not None, self.serial is not None]
        )



from __future__ import annotations

from datetime import datetime

from pydantic import Field, computed_field

from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.models import EnumModel, Solaredge2MQTTBaseModel
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantSensorType as HASensor,
)
from solaredge2mqtt.services.models import Component


class HistoricBaseModel(Component):
    SOURCE = "energy"

    info: HistoricInfo

    def __init__(self, data: dict, period: HistoricPeriod, **kwargs):
        super().__init__(
            info=HistoricInfo(
                unit=data.get("unit", None),
                period=period,
                start=data["_start"],
                stop=data["_stop"]
            ),
            **kwargs,
        )

    def mqtt_topic(self) -> str:
        parts = [self.SOURCE]

        if self.info.unit:
            parts.append(self.info.unit)

        parts.append(self.info.period.topic)

        return "/".join(parts)

    def __str__(self) -> str:
        return f"{self.SOURCE}: {self.info.period}"


class HistoricEnergy(HistoricBaseModel):
    pv_production: float = Field(
        **HASensor.ENERGY_KWH.field("PV production"))
    inverter: InverterEnergy = Field(title="Inverter")
    grid: GridEnergy = Field(title="Grid")
    battery: BatteryEnergy = Field(title="Battery")
    consumer: ConsumerEnergy = Field(title="Consumer")
    money: HistoricMoney | None = Field(None, title="Money")

    def __init__(
        self,
        energy_data: dict,
        period: HistoricPeriod,
    ):
        logger.trace(energy_data)
        pv_production = energy_data["pv_production"]

        subclass_values: dict[str, dict[str, float]] = {}

        for key, value in energy_data.items():
            keys = key.split("_")
            if len(keys) < 2:
                continue

            if keys[0] in ["consumer", "inverter", "grid", "battery", "money"]:
                if keys[0] not in subclass_values:
                    subclass_values[keys[0]] = {}

                subclass_values[keys[0]]["_".join(keys[1:])] = round(value, 3)

        logger.debug(subclass_values["consumer"])

        super().__init__(
            period=period,
            data=energy_data,
            pv_production=round(pv_production, 3),
            inverter=InverterEnergy(**subclass_values["inverter"]),
            grid=GridEnergy(**subclass_values["grid"]),
            battery=BatteryEnergy(**subclass_values["battery"]),
            consumer=ConsumerEnergy(**subclass_values["consumer"]),
            money=(
                HistoricMoney(**subclass_values["money"])
                if "money" in subclass_values
                else None
            ),
        )

    @computed_field(title="Self consumption rate")
    @property
    def self_consumption_rates(self) -> SelfConsumptionRate:
        return SelfConsumptionRate(self)

    @computed_field(title="Self sufficiency rate")
    @property
    def self_sufficiency_rates(self) -> SelfSufficiencyRate:
        return SelfSufficiencyRate(self)

    @property
    def has_unit(self) -> bool:
        return self.unit is not None

    def homeassistant_device_info(self) -> dict[str, any]:
        parts = ["Energy"]

        if self.info.unit:
            parts.append(self.info.unit)

        parts.append(self.info.period.title)

        return self._default_homeassistant_device_info(
            " - ".join(parts),
        )


class HistoricMoney(Solaredge2MQTTBaseModel):
    delivered: float = Field(**HASensor.MONETARY.field("Delivered"))
    saved: float = Field(**HASensor.MONETARY.field("Saved"))
    consumed: float = Field(**HASensor.MONETARY.field("Consumed"))
    price_in: float = Field(exclude=True)
    price_out: float = Field(exclude=True)

    @computed_field(**HASensor.MONETARY_BALANCE.field("Balance grid"))
    @property
    def balance_grid(self) -> float:
        return round(self.delivered - self.consumed, 3)

    @computed_field(**HASensor.MONETARY_BALANCE.field("Balance total"))
    @property
    def balance_total(self) -> float:
        return round(self.balance_grid + self.saved, 3)

    @computed_field(**HASensor.MONETARY.field("Total revenue"))
    @property
    def total_revenue(self) -> float:
        return round(self.delivered + self.saved, 3)


class HistoricInfo(Solaredge2MQTTBaseModel):
    unit: str | None = Field(None)
    period: HistoricPeriod
    start: datetime
    stop: datetime


class HistoricQuery(EnumModel):
    ACTUAL = "actual_unit"
    LAST = "historic_unit"

    def __init__(self, query: str) -> None:
        self._query: str = query

    @property
    def query(self) -> str:
        return self._query


class HistoricPeriod(EnumModel):
    LAST_HOUR = (
        "last_hour",
        "Last hour",
        "1h",
        HistoricQuery.LAST,
        True,
    )
    TODAY = "today", "Today", "1d", HistoricQuery.ACTUAL, True
    YESTERDAY = (
        "yesterday",
        "Yesterday",
        "1d",
        HistoricQuery.LAST,
        True,
    )
    THIS_WEEK = "this_week", "This week", "1w", HistoricQuery.ACTUAL, False
    LAST_WEEK = "last_week", "Last week", "1w", HistoricQuery.LAST, False
    THIS_MONTH = (
        "this_month",
        "This month",
        "1mo",
        HistoricQuery.ACTUAL,
        True,
    )
    LAST_MONTH = "last_month", "Last month", "1mo", HistoricQuery.LAST, True
    THIS_YEAR = (
        "this_year",
        "This year",
        "1y",
        HistoricQuery.ACTUAL,
        True,
    )
    LAST_YEAR = "last_year", "Last year", "1y", HistoricQuery.LAST, True
    LIFETIME = (
        "lifetime",
        "Lifetime",
        "99y",
        HistoricQuery.ACTUAL,
        True,
    )

    def __init__(
        self,
        topic: str,
        title: str,
        unit: str,
        query: HistoricQuery,
        auto_discovery: bool,
    ) -> None:
        self._topic: str = topic
        self._title: str = title
        self._unit: str = unit
        self._query: HistoricQuery = query
        self._auto_discovery: bool = auto_discovery

    @property
    def topic(self) -> str:
        return self._topic

    @property
    def title(self) -> str:
        return self._title

    @property
    def unit(self) -> str:
        return self._unit

    @property
    def query(self) -> HistoricQuery:
        return self._query

    @property
    def auto_discovery(self) -> bool:
        return self._auto_discovery


class InverterEnergy(Solaredge2MQTTBaseModel):
    production: float = Field(**HASensor.ENERGY_KWH.field("Production"))
    consumption: float = Field(**HASensor.ENERGY_KWH.field("Consumption"))
    dc_power: float = Field(**HASensor.POWER_W.field("DC production"))
    pv_production: float = Field(
        **HASensor.ENERGY_KWH.field("PV production"))
    battery_production: float = Field(
        **HASensor.ENERGY_KWH.field("Battery production")
    )


class GridEnergy(Solaredge2MQTTBaseModel):
    delivery: float = Field(**HASensor.ENERGY_KWH.field("Delivery"))
    consumption: float = Field(**HASensor.ENERGY_KWH.field("Consumption"))


class BatteryEnergy(Solaredge2MQTTBaseModel):
    charge: float = Field(**HASensor.ENERGY_KWH.field("Charge"))
    discharge: float = Field(**HASensor.ENERGY_KWH.field("Discharge"))


class ConsumerEnergy(Solaredge2MQTTBaseModel):
    house: float = Field(**HASensor.ENERGY_KWH.field("House"))
    evcharger: float = Field(**HASensor.ENERGY_KWH.field("EV-Charger"))
    inverter: float = Field(**HASensor.ENERGY_KWH.field("Inverter"))

    total: float = Field(**HASensor.ENERGY_KWH.field("Total"))

    used_production: float = Field(
        **HASensor.ENERGY_KWH.field("Used production"))
    used_pv_production: float = Field(
        **HASensor.ENERGY_KWH.field("Used PV production")
    )
    used_battery_production: float = Field(
        **HASensor.ENERGY_KWH.field("Used battery production")
    )


class SelfConsumptionRate(Solaredge2MQTTBaseModel):
    grid: int = Field(**HASensor.PERCENTAGE.field("Grid"))
    battery: int = Field(**HASensor.PERCENTAGE.field("Battery"))
    pv: int = Field(**HASensor.PERCENTAGE.field("PV"))
    total: int = Field(**HASensor.PERCENTAGE.field("Total"))

    def __init__(self, energy: HistoricEnergy):
        if energy.inverter.production > 0:
            grid_rate = int(
                round(energy.grid.delivery / energy.inverter.production * 100)
            )
            battery_rate = int(
                round(
                    energy.consumer.used_battery_production
                    / energy.inverter.production
                    * 100
                )
            )
            pv_rate = 100 - grid_rate - battery_rate
            total = battery_rate + pv_rate
        else:
            grid_rate = 0
            battery_rate = 0
            pv_rate = 0
            total = 0

        super().__init__(
            grid=grid_rate,
            battery=battery_rate,
            pv=pv_rate,
            total=total,
        )


class SelfSufficiencyRate(Solaredge2MQTTBaseModel):
    grid: int = Field(**HASensor.PERCENTAGE.field("Grid"))
    battery: int = Field(**HASensor.PERCENTAGE.field("Battery"))
    pv: int = Field(**HASensor.PERCENTAGE.field("PV"))
    total: int = Field(**HASensor.PERCENTAGE.field("Total"))

    def __init__(self, energy: HistoricEnergy):
        if energy.consumer.total > 0:
            grid_rate = int(
                round(energy.grid.consumption / energy.consumer.total * 100)
            )
            battery_rate = int(
                round(
                    energy.consumer.used_battery_production
                    / energy.consumer.total
                    * 100
                )
            )
            pv_rate = 100 - grid_rate - battery_rate
            total = battery_rate + pv_rate
        else:
            grid_rate = 0
            battery_rate = 0
            pv_rate = 0
            total = 0

        super().__init__(
            grid=grid_rate,
            battery=battery_rate,
            pv=pv_rate,
            total=total,
        )



from pydantic import BaseModel, Field, SecretStr


class WeatherSettings(BaseModel):
    api_key: SecretStr | None = Field(None)
    language: str = Field("en")
    retain: bool = Field(False)

    @property
    def is_configured(self) -> bool:
        return self.api_key is not None



import json
from pathlib import Path

from setuptools_scm import get_version


def generate_version_json():
    version = get_version(root=".", relative_to=__file__)
    data = {"version": version}
    path = Path("solaredge2mqtt/version.json")
    with path.open("w", encoding="utf-8") as f:
        json.dump(data, f, indent=2)
        f.write("\n")  # sauberes Newline am Ende der Datei
    print(f"Generated {path} with version {version}")


if __name__ == "__main__":
    generate_version_json()



from __future__ import annotations

from typing import TYPE_CHECKING

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.core.influxdb import InfluxDBAsync
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.core.timer.events import IntervalBaseTriggerEvent
from solaredge2mqtt.services.modbus import Modbus
from solaredge2mqtt.services.modbus.models.battery import ModbusBattery
from solaredge2mqtt.services.powerflow.events import PowerflowGeneratedEvent
from solaredge2mqtt.services.powerflow.models import Powerflow
from solaredge2mqtt.services.wallbox import WallboxClient

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings.models import ServiceSettings


class PowerflowService:

    def __init__(
        self,
        settings: ServiceSettings,
        event_bus: EventBus,
        influxdb: InfluxDBAsync | None = None,
    ):
        self.settings = settings

        self.influxdb = influxdb

        self.modbus = Modbus(self.settings, event_bus)

        self.wallbox = (
            WallboxClient(self.settings.wallbox, event_bus)
            if self.settings.is_wallbox_configured
            else None
        )

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(
            IntervalBaseTriggerEvent, self.calculate_powerflow)

    async def async_init(self) -> None:
        await self.modbus.async_init()

    async def calculate_powerflow(self, _) -> None:
        units = await self.modbus.get_data()

        if "leader" not in units:
            raise InvalidDataException("Invalid modbus data")

        batteries = {
            f"{unit_key}:{battery_key}": battery
            for unit_key, unit in units.items()
            for battery_key, battery in unit.batteries.items()
        }

        for battery in batteries.values():
            if not battery.is_valid:
                logger.debug(battery)
                raise InvalidDataException("Invalid battery data")

        evcharger = 0
        wallbox_data = None
        if self.settings.is_wallbox_configured:
            try:
                wallbox_data = await self.wallbox.get_data()
                logger.trace(
                    "Wallbox: {wallbox_data.power} W", wallbox_data=wallbox_data
                )
                evcharger = wallbox_data.power
            except ConfigurationException as ex:
                logger.warning(f"{ex.component}: {ex.message}")

        powerflows: dict[str, Powerflow] = {}

        for unit_key, unit in units.items():
            if unit_key == "leader":
                powerflows[unit_key] = Powerflow.from_modbus(unit, evcharger)
            else:
                powerflows[unit_key] = Powerflow.from_modbus(unit)

            logger.trace(powerflows[unit_key].model_dump_json())

        if self.settings.modbus.has_followers:
            powerflow = Powerflow.cumulated_powerflow(powerflows)
            powerflows["cumulated"] = powerflow
        else:
            powerflow = powerflows["leader"]

        if not powerflow.is_valid(self.settings.powerflow.external_production):
            logger.info(powerflow)
            raise InvalidDataException("Invalid powerflow data")

        if Powerflow.is_not_valid_with_last(powerflow):
            logger.debug(powerflow)
            raise InvalidDataException(
                "Value change not valid, skipping this loop")

        await self.write_to_influxdb(powerflows, batteries)

        logger.debug(powerflow)
        logger.info(
            "Powerflow: PV {pv_production} W, Inverter {inverter.power} W, "
            + "House {consumer.house} W, Grid {grid.power} W, "
            + "Battery {battery.power} W, Wallbox {consumer.evcharger} W",
            pv_production=powerflow.pv_production,
            inverter=powerflow.inverter,
            consumer=powerflow.consumer,
            grid=powerflow.grid,
            battery=powerflow.battery,
        )

        await self.publish_modbus(units)

        await self.publish_wallbox(wallbox_data)

        await self.publish_powerflow(powerflows)

    async def publish_modbus(self, units):
        for key, unit in units.items():
            await self.event_bus.emit(
                MQTTPublishEvent(
                    unit.inverter.mqtt_topic(
                        self.settings.modbus.has_followers),
                    unit.inverter,
                    self.settings.modbus.retain)
            )

            for key, component in {**unit.meters, **unit.batteries}.items():
                await self.event_bus.emit(
                    MQTTPublishEvent(
                        f"{component.mqtt_topic(self.settings.modbus.has_followers)}/{key.lower()}",
                        component,
                        self.settings.modbus.retain
                    )
                )

    async def publish_wallbox(self, wallbox_data):
        if wallbox_data is not None:
            await self.event_bus.emit(
                MQTTPublishEvent(
                    wallbox_data.mqtt_topic(),
                    wallbox_data,
                    self.settings.wallbox.retain
                )
            )

    async def publish_powerflow(self, powerflows: dict[str, Powerflow]) -> None:
        if self.settings.modbus.has_followers:
            for pf in powerflows.values():
                await self.event_bus.emit(
                    MQTTPublishEvent(
                        pf.mqtt_topic(),
                        pf,
                        self.settings.powerflow.retain
                    )
                )
        else:
            powerflow = powerflows["leader"]
            await self.event_bus.emit(
                MQTTPublishEvent(
                    powerflow.mqtt_topic(),
                    powerflow,
                    self.settings.powerflow.retain
                )
            )

        await self.event_bus.emit(PowerflowGeneratedEvent(powerflows))

    async def write_to_influxdb(
        self,
        powerflows: dict[str, Powerflow],
        batteries_data: dict[str, ModbusBattery],
    ):
        if self.influxdb is not None:
            points = []

            for powerflow in powerflows.values():
                points.append(powerflow.prepare_point())

            for battery in batteries_data.values():
                points.append(battery.prepare_point())

            await self.influxdb.write_points(points)

    async def close(self) -> None:
        if self.settings.is_wallbox_configured:
            await self.wallbox.close()



from solaredge2mqtt.services.modbus.sunspec.base import SunSpecOffset, SunSpecRegister
from solaredge2mqtt.services.modbus.sunspec.values import SunSpecValueType


class SunSpecBatteryInfoRegister(SunSpecRegister):
    C_MANUFACTURER = "c_manufacturer", 57600, SunSpecValueType.STRING, True, 16
    C_MODEL = "c_model", 57616, SunSpecValueType.STRING, True, 16
    C_VERSION = "c_version", 57632, SunSpecValueType.STRING, True, 16
    C_SERIALNUMBER = "c_serialnumber", 57648, SunSpecValueType.STRING, True, 16
    C_DEVICEADDRESS = "c_deviceaddress", 57664, SunSpecValueType.UINT16, True
    C_SUNSPEC_DID = "c_sunspec_did", 57665, SunSpecValueType.UINT16, True


class SunSpecBatteryRegister(SunSpecRegister):
    RATED_ENERGY = "rated_energy", 57666, SunSpecValueType.FLOAT32
    MAXIMUM_CHARGE_CONTINUOUS_POWER = (
        "maximum_charge_continuous_power",
        57668,
        SunSpecValueType.FLOAT32,
    )
    MAXIMUM_DISCHARGE_CONTINUOUS_POWER = (
        "maximum_discharge_continuous_power",
        57670,
        SunSpecValueType.FLOAT32,
    )
    MAXIMUM_CHARGE_PEAK_POWER = (
        "maximum_charge_peak_power",
        57672,
        SunSpecValueType.FLOAT32,
    )
    MAXIMUM_DISCHARGE_PEAK_POWER = (
        "maximum_discharge_peak_power",
        57674,
        SunSpecValueType.FLOAT32,
    )

    AVERAGE_TEMPERATURE = "average_temperature", 57708, SunSpecValueType.FLOAT32
    MAXIMUM_TEMPERATURE = "maximum_temperature", 57710, SunSpecValueType.FLOAT32

    INSTANTANEOUS_VOLTAGE = (
        "instantaneous_voltage",
        57712,
        SunSpecValueType.FLOAT32,
        True,
    )
    INSTANTANEOUS_CURRENT = (
        "instantaneous_current",
        57714,
        SunSpecValueType.FLOAT32,
        True,
    )
    INSTANTANEOUS_POWER = "instantaneous_power", 57716, SunSpecValueType.FLOAT32, True

    LIFETIME_EXPORT_ENERGY_COUNTER = (
        "lifetime_export_energy_counter",
        57718,
        SunSpecValueType.UINT64,
    )
    LIFETIME_IMPORT_ENERGY_COUNTER = (
        "lifetime_import_energy_counter",
        57722,
        SunSpecValueType.UINT64,
    )

    MAXIMUM_ENERGY = "maximum_energy", 57726, SunSpecValueType.FLOAT32
    AVAILABLE_ENERGY = "available_energy", 57728, SunSpecValueType.FLOAT32

    SOH = "soh", 57730, SunSpecValueType.FLOAT32, True
    SOE = "soe", 57732, SunSpecValueType.FLOAT32, True

    STATUS = "status", 57734, SunSpecValueType.UINT32, True
    STATUS_INTERNAL = "status_internal", 57736, SunSpecValueType.UINT32

    EVENT_LOG = "event_log", 57738, SunSpecValueType.UINT32
    EVENT_LOG_INTERNAL = "event_log_internal", 57746, SunSpecValueType.UINT32

    @staticmethod
    def wordorder() -> str:
        return "little"


class SunSpecBatteryOffset(SunSpecOffset):
    BATTERY0 = "battery0", 0
    BATTERY1 = "battery1", 256



from pydantic import BaseModel, Field


class PriceSettings(BaseModel):
    consumption: float = Field(None)
    delivery: float = Field(None)
    currency: str = Field(None)

    @property
    def is_configured(self) -> bool:
        return self.is_consumption_configured or self.is_delivery_configured

    @property
    def is_consumption_configured(self) -> bool:
        return self.consumption is not None and self.currency is not None

    @property
    def is_delivery_configured(self) -> bool:
        return self.delivery is not None and self.currency is not None

    @property
    def price_in(self) -> float:
        return self.consumption or 0.0

    @property
    def price_out(self) -> float:
        return self.delivery or 0.0


class EnergySettings(BaseModel):
    retain: bool = Field(False)



import sys

from loguru import logger

from solaredge2mqtt.core.logging.models import LoggingLevelEnum


def initialize_logging(logging_level: LoggingLevelEnum) -> None:
    logger.configure(
        handlers=[{
            "sink": sys.stdout,
            "level": logging_level.level
        }])



from pydantic import BaseModel, Field, SecretStr


class MonitoringSettings(BaseModel):
    site_id: str = Field(None)
    username: str = Field(None)
    password: SecretStr = Field(None)
    retain: bool = Field(False)

    @property
    def is_configured(self) -> bool:
        return all(
            [
                self.site_id is not None,
                self.username is not None,
                self.password is not None,
            ]
        )



import asyncio
import time

import jwt
from aiohttp.client_exceptions import ClientResponseError
from pydantic import BaseModel, Field

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.services.http_async import HTTPClientAsync
from solaredge2mqtt.services.wallbox.events import WallboxReadEvent
from solaredge2mqtt.services.wallbox.models import WallboxAPI
from solaredge2mqtt.services.wallbox.settings import WallboxSettings

LOGIN_URL = "https://{host}:8443/v2/jwt/login"
REFRESH_URL = "https://{host}:8443/v2/jwt/refresh"
WALLBOX_URL = "https://{host}:8443/v2/wallboxes/{serial}"


class AuthorizationTokens(BaseModel):
    access_token: str | None = Field(None, alias="accessToken")
    refresh_token: str | None = Field(None, alias="refreshToken")

    @property
    def access_token_expires(self) -> int:
        expires = self.get_exp_claim(self.access_token)
        logger.debug("Access token expires at: {expires}", expires=expires)
        return expires

    @property
    def refresh_token_expires(self) -> int:
        expires = self.get_exp_claim(self.refresh_token)
        logger.debug("Refresh token expires at: {expires}", expires=expires)
        return expires

    @staticmethod
    def get_exp_claim(token: str) -> int:
        try:
            # ⚠️ This only extracts the 'exp' claim without verifying the JWT signature.
            # It's used for internal expiration checks, not for authentication
            # or security decisions.
            payload = jwt.decode(
                token, options={"verify_signature": False})  # NOSONAR
            return payload["exp"]
        except Exception as e:
            logger.warning(
                "Failed to decode JWT for exp claim: {error}", error=e)
            raise InvalidDataException("Cannot read token expiration") from e


class WallboxClient(HTTPClientAsync):
    def __init__(self, settings: WallboxSettings, event_bus: EventBus):
        super().__init__("Wallbox API")
        self.settings = settings
        self.event_bus = event_bus
        self.authorization: AuthorizationTokens | None = None

        logger.info(
            "Using Wallbox charger: {host}",
            host=settings.host,
        )

    async def get_data(self) -> WallboxAPI | None:
        wallbox = None
        try:
            await self._get_access()

            async with asyncio.timeout(5):
                response = await self._get(
                    WALLBOX_URL.format(
                        host=self.settings.host, serial=self.settings.serial
                    ),
                    headers={
                        "Authorization": f"Bearer {self.authorization.access_token}"},
                    verify=False,
                    login=self.login,
                )

            if response is None:
                raise InvalidDataException("Invalid Wallbox data")

            wallbox = WallboxAPI(response)
            logger.info(
                f"Wallbox: {wallbox.state}, {wallbox.power / 1000} kW",
                wallbox=wallbox,
            )

            await self.event_bus.emit(WallboxReadEvent(wallbox))
        except (ClientResponseError, asyncio.TimeoutError) as error:
            raise InvalidDataException(
                f"Cannot read Wallbox data: {error}") from error

        return wallbox

    async def _get_access(self) -> None:
        current_timestamp = int(time.time())

        if self.authorization is None:
            await self.login()
        elif self.authorization.access_token_expires < current_timestamp + 60:
            if self.authorization.refresh_token_expires < current_timestamp + 60:
                await self.login()
            else:
                await self._refresh_token()
        else:
            logger.debug("Wallbox access token still valid")

    async def login(self):
        try:
            logger.info("Logging in to Wallbox charger...")
            self.authorization = None
            async with asyncio.timeout(5):
                response = await self._post(
                    LOGIN_URL.format(host=self.settings.host),
                    json={
                        "password": self.settings.password.get_secret_value(),
                        "username": "admin",
                    },
                    verify=False,
                )

            if response is None:
                raise ConfigurationException(
                    "wallbox", "Invalid Wallbox login")

            self.authorization = AuthorizationTokens(**response)
            logger.info("Logged in to EV charger")
        except (ClientResponseError, asyncio.TimeoutError) as error:
            raise ConfigurationException(
                "wallbox", "Unable to login to EV charger"
            ) from error

    async def _refresh_token(self):
        logger.info("Refreshing access token Wallbox...")
        async with asyncio.timeout(5):
            response = await self._post(
                REFRESH_URL.format(host=self.settings.host),
                headers={
                    "Authorization": f"Bearer {self.authorization.refresh_token}"},
                verify=False,
                login=self.login,
            )

        self.authorization.access_token = response["accessToken"]
        logger.info("Refreshed access token Wallbox")



import asyncio
import json
from datetime import datetime, timezone

from aiohttp import ClientResponseError

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.core.influxdb import InfluxDBAsync, Point
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.core.timer.events import Interval15MinTriggerEvent
from solaredge2mqtt.services.http_async import HTTPClientAsync
from solaredge2mqtt.services.monitoring.models import (
    LogicalInfo,
    LogicalInverter,
    LogicalModule,
    LogicalString,
)
from solaredge2mqtt.services.monitoring.settings import MonitoringSettings

LOGIN_URL = "https://monitoring.solaredge.com/solaredge-apigw/api/login"
LOGICAL_URL = "https://monitoring.solaredge.com/solaredge-apigw/api/sites/{site_id}/layout/logical"
POWER_PUBLIC_URL = "https://monitoring.solaredge.com/solaredge-web/p/playbackData"
CONTENT_TYPE_FORM_URLENCODED = "application/x-www-form-urlencoded"


class MonitoringSite(HTTPClientAsync):
    def __init__(
        self,
        settings: MonitoringSettings,
        event_bus: EventBus,
        influxdb: InfluxDBAsync | None,
    ) -> None:
        super().__init__("Monitoring Site")
        self.settings = settings

        self.influxdb: InfluxDBAsync | None = influxdb

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(Interval15MinTriggerEvent, self.get_data)

    async def get_data(self, _):
        energies = await self.get_modules_energy()
        powers = await self.get_modules_power()

        modules = self.merge_modules(energies, powers)

        energy_total = 0
        count_modules = 0

        await self.save_to_influxdb(modules)
        await self.publish_mqtt(modules, energy_total, count_modules)

    async def get_modules_energy(self) -> dict[str, LogicalModule]:
        logical = await self._get_logical()

        if logical is None:
            raise InvalidDataException(
                "Unable to read logical layout from monitoring site"
            )

        inverters = self._parse_inverters(
            logical["logicalTree"]["children"], logical["reportersData"]
        )

        modules = {}

        for inverter in inverters:
            logger.debug(
                "Inverter: {inverter}", inverter=inverter.model_dump_json(indent=4)
            )
            for string in inverter.strings:
                for module in string.modules:
                    modules[module.info.identifier] = module

        return modules

    async def _get_logical(self) -> dict | None:
        result = None
        try:
            if not self.cookie_exists("CSRF-TOKEN"):
                await self.login()

            async with asyncio.timeout(10):
                result = await self._get(
                    LOGICAL_URL.format(site_id=self.settings.site_id),
                    headers={
                        "Content-Type": CONTENT_TYPE_FORM_URLENCODED,
                        "X-CSRF-TOKEN": self.get_cookie("CSRF-TOKEN"),
                    }
                )
        except (ClientResponseError, asyncio.TimeoutError) as error:
            raise InvalidDataException(
                "Unable to read logical layout") from error

        return result

    def _parse_inverters(self, inverter_objs, reporters_data) -> list[LogicalInverter]:
        inverters = []

        for inverter_obj in inverter_objs:
            info = LogicalInfo.map(inverter_obj["data"])
            if "INVERTER" in info["type"]:
                inverter = LogicalInverter(
                    info=info,
                    energy=(
                        reporters_data[info["identifier"]]["unscaledEnergy"]
                        if info["identifier"] in reporters_data
                        else None
                    ),
                )

                self._parse_strings(
                    inverter, inverter_obj["children"], reporters_data)

                inverters.append(inverter)

            else:
                logger.info("Unknown inverter type: {type}", type=info["type"])

        return inverters

    def _parse_strings(self, inverter, string_objs, reporters_data):
        for string_obj in string_objs:
            info = LogicalInfo.map(string_obj["data"])
            string = LogicalString(
                info=info,
                energy=(
                    reporters_data[info["identifier"]]["unscaledEnergy"]
                    if info["identifier"] in reporters_data
                    else None
                ),
            )

            self._parse_panels(string, string_obj["children"], reporters_data)

            inverter.strings.append(string)

    def _parse_panels(self, string, panel_objs, reporters_data):
        for panel_obj in panel_objs:
            info = LogicalInfo.map(panel_obj["data"])
            panel = LogicalModule(
                info=info,
                energy=(
                    reporters_data[info["identifier"]]["unscaledEnergy"]
                    if info["identifier"] in reporters_data
                    else None
                ),
            )

            string.modules.append(panel)

    async def get_modules_power(self) -> dict[str, dict[datetime, float]]:
        playback = await self._get_playback()

        if playback is None:
            raise InvalidDataException(
                "Unable to read playback data from monitoring site"
            )

        modules = {}

        for date_str, reporters_data in playback["reportersData"].items():
            date = datetime.strptime(
                date_str, "%a %b %d %H:%M:%S GMT %Y").astimezone()

            for entries in reporters_data.values():
                for entry in entries:
                    key = entry["key"]
                    if key not in modules:
                        modules[key] = {}

                    modules[key][date] = float(entry["value"])

        logger.debug(modules)

        return modules

    async def _get_playback(self) -> dict | None:
        result = None
        try:
            if not self.cookie_exists("CSRF-TOKEN"):
                await self.login()
            async with asyncio.timeout(10):
                playback_data = await self._post(
                    POWER_PUBLIC_URL,
                    data={
                        "fieldId": self.settings.site_id,
                        "timeUnit": 4,
                        "CSRF": self.get_cookie("CSRF-TOKEN"),
                    },
                    headers={
                        "Content-Type": CONTENT_TYPE_FORM_URLENCODED,
                        "X-CSRF-TOKEN": self.get_cookie("CSRF-TOKEN"),
                    },
                    expect_json=False,
                )

            response = (
                playback_data.replace("'", '"')
                .replace("Array", "")
                .replace("key", '"key"')
                .replace("value", '"value"')
                .replace("timeUnit", '"timeUnit"')
                .replace("fieldData", '"fieldData"')
                .replace("reportersData", '"reportersData"')
            )

            result = json.loads(response)
        except (ClientResponseError, asyncio.TimeoutError) as error:
            raise InvalidDataException(
                "Unable to read logical layout") from error

        return result

    async def login(self) -> None:
        try:
            async with asyncio.timeout(10):
                await self._post(
                    LOGIN_URL,
                    headers={"Content-Type": CONTENT_TYPE_FORM_URLENCODED},
                    data={
                        "j_username": self.settings.username,
                        "j_password": self.settings.password.get_secret_value(),
                    },
                    expect_json=False,
                )

            logger.info("Login to monitoring site successful")
        except (ClientResponseError, asyncio.TimeoutError) as error:
            raise ConfigurationException(
                "Monitoring", "Unable to login to monitoring account"
            ) from error

    @staticmethod
    def merge_modules(
        energies: dict[str, LogicalModule], powers: dict[str, dict[datetime, float]]
    ) -> dict[str, LogicalModule]:
        modules = {}

        for sid, module in energies.items():
            if sid in powers:
                module.power = powers[sid]

            modules[sid] = module

        return modules

    async def save_to_influxdb(self, modules):
        if self.influxdb is not None:
            points = []
            for module in modules.values():
                if module.power is not None:
                    for date, module_power in module.power.items():
                        point = Point("modules")
                        point.field("power", module_power)
                        point.time(date.astimezone(timezone.utc))
                        point.tag("serialnumber", module.info.serialnumber)
                        point.tag("name", module.info.name)
                        point.tag("identifier", module.info.identifier)
                        points.append(point)

            await self.influxdb.write_points(points)

    async def publish_mqtt(self, modules, energy_total, count_modules):
        for module in modules.values():
            if module.energy is not None:
                count_modules += 1
                energy_total += module.energy

            await self.event_bus.emit(
                MQTTPublishEvent(
                    f"monitoring/module/{module.info.serialnumber}",
                    module,
                    self.settings.retain
                )
            )

        logger.info(
            "Read from monitoring total energy: {energy_total} kWh "
            "from {count_modules} modules",
            energy_total=energy_total / 1000,
            count_modules=count_modules,
        )

        await self.event_bus.emit(
            MQTTPublishEvent(
                "monitoring/pv_energy_today",
                energy_total,
                self.settings.retain,
            )
        )



import logging

from solaredge2mqtt.core.models import EnumModel


class LoggingLevelEnum(EnumModel):
    DEBUG = "DEBUG", logging.DEBUG
    INFO = "INFO", logging.INFO
    WARNING = "WARNING", logging.WARNING
    ERROR = "ERROR", logging.ERROR
    CRITICAL = "CRITICAL", logging.CRITICAL

    def __init__(self, description: str, level: int):
        # pylint: disable=super-init-not-called
        self._description: str = description
        self._level: int = level

    @property
    def description(self) -> str:
        return self._description

    @property
    def level(self) -> int:
        return self._level
